#!/bin/sh

exec </dev/console
exec >/dev/console
exec 2>/dev/console

source /etc/env

ZIPFILE="${1}"
SAVE="/rw/etc.tar"

SAM3U_IMG="sam3u.fbin"
UBOOT_IMG="u-boot-qspi.bin"
FPGADATA_IMG="top-part.bin"
KERNEL_IMG="uImage.itb"
MACHXO_IMG="machxo.jed"

ROOTFS="rootfs.squashfs"
ROOTFS_TARGET="rootfs.squashfs"
OPTFS="optfs"
SUBFS="subfs"
DATAFS="varfs.tgz"
REMOVELIST="remove.list"

STATUSDIR="/rw/var/instrument"
STATUSOLD="${STATUSDIR}/update_oldfail"
STATUSERR="${STATUSDIR}/update_failure"
STATUSPOW="${STATUSDIR}/update_powercut"

RSUDESC="/var/instrument/rsu_description.txt"
NRPMALIST="/rw/var/instrument/update_nrpma.txt"

LED="/sys/class/leds/nrpz::user"

# start_led()
start_led()
{
	# 80% white
	echo none >${LED}/trigger
	printf "%d" 0xCCCCCC >${LED}/brightness
}

# blink_led()
blink_led()
{
	# 1Hz blink (default)
	echo timer >${LED}/trigger
}

# panic(msg)
panic()
{
	echo "${1}" >&2
	echo "${1}" >"${STATUSERR}"
	rm -f "${STATUSPOW}"
	umount /rw
	reboot -f
	sleep 10
}

# panic_space(path)
panic_space()
{
	rm -rf "${1}"
	panic "Writing ${1} to storage failed!"
}

# file_length(path)
file_length()
{
	stat -c '%s' ${1} 2>/dev/null || echo -n -1
}

# zip_list(file)
zip_list()
{
	unzip -l ${ZIPFILE} "${1}" || panic "Update file corrupt (${1} unzip)."
}

# zip_length(file)
zip_length()
{
	echo " $(zip_list "${1}" | grep "\<${1}$")" | tr -s ' ' | cut -f2 -d' '
}

# zip_read(file)
zip_read()
{
	unzip -p ${ZIPFILE} "${1}" || panic "Update file corrupt (${1} unzip)."
}

# zip_extract(path,file,[len])
zip_extract()
{
	local MD5SUM

	[ -z "${3}" ] && return
	if ! unzip -p ${ZIPFILE} "${2}" > "${1}" ; then
		rm -f "${1}"
		panic "Writing ${1} failed (${2} unzip)!"
	fi
	MD5SUM=$(md5sum ${1} | tr -s ' ' | cut -f1 -d' ')
	if [ "${MD5SUM}" != "$(unzip -p ${ZIPFILE} ${2}.md5sum)" ] ; then
		rm -f "${1}"
		panic "Writing ${1} failed (${2} md5sum)!"
	fi
}

# md5sum_cmp(path,file,len)
md5sum_cmp()
{
	[ ! -e ${1} ] && return 1
	[ $(file_length ${1}) -ne ${3} ]  && return 1
	test "$(md5sum ${1} | tr -s ' ' | cut -f1 -d' ')" = "$(zip_read ${2}.md5sum)"
	return $?
}

# check_md5sum(file,len)
check_md5sum()
{
	local MD5SUM

	if [ "${2}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			panic "Update file corrupt (${1} md5sum)."
		fi
	fi
}

# get_mtd_size(mtdno)
get_mtd_size()
{
	local SIZE

	SIZE=$(grep "^mtd${1}:" /proc/mtd | cut -f2 -d' ')
	if [ -z "${SIZE}" ]; then
		echo 0
	else
		printf "%d\n" 0x${SIZE}
	fi
}

# get_mtd_erase()
get_mtd_erase()
{
	printf "%d\n" 0x$(cat /proc/mtd | grep "^mtd0:" | cut -f3 -d' ')
}

# mtd_read(mtdno,len)
mtd_read()
{
	local MTD_ERASE=$(get_mtd_erase)
	local SIZE=${2}
	local BLOCKS=$(( ${SIZE} / ${MTD_ERASE} ))

	# Note: "dd if=/dev/mtd${1} bs=${2} count=1" would be easier, 
	# but BusyBox dd uses malloc(ibs) ...
	dd if=/dev/mtd${1} bs=${MTD_ERASE} count=${BLOCKS} 2>/dev/null
	SIZE=$(( ${SIZE} % ${MTD_ERASE} ))
	dd if=/dev/mtd${1} bs=${MTD_ERASE} count=1 skip=${BLOCKS} 2>/dev/null |
	dd bs=${SIZE} count=1 2>/dev/null
}

# check_image(file,mtdno,[len])
check_image()
{
	local TMP_IMAGE=/tmp/image
	local MD5SUM
	local ret

	if [ -z "${3}" ]; then
		return 1
	fi

	if [ ${3} -gt $(get_mtd_size ${2}) ]; then
		panic "Image ${1} too big for partition."
	fi

	MD5SUM=$(zip_read ${1} | tee "${TMP_IMAGE}" | md5sum | tr -s ' ' | cut -f1 -d' ')
	if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
		panic "Update file corrupt (${1} md5sum)."
	fi

	mtd_read ${2} ${3} | cmp -s ${TMP_IMAGE}
	ret=$?
	rm -f ${TMP_IMAGE}
	return ${ret}
}

# update_image(file,mtdno,[len])
update_image()
{
	local TMP_IMAGE=/tmp/image
	local MTD_ERASE=$(get_mtd_erase)
	local SIZE=${3:-0}
	local BLOCKS=$(( ( ${SIZE} + ${MTD_ERASE} - 1 ) / ${MTD_ERASE} ))

	if [ -z "${3}" ]; then
		return
	fi

	echo "Updating ${1} image..."
	zip_extract ${TMP_IMAGE} ${1} ${3}

	flash_erase -q /dev/mtd${2} 0 ${BLOCKS}
	if [ $? -ne 0 ]; then
		panic "Erasing /dev/mtd${2} failed!"
	fi

	dd if=${TMP_IMAGE} of=/dev/mtd${2} bs=${MTD_ERASE}
	if [ $? -ne 0 ]; then
		panic "Writing ${1} image failed!"
	fi

	mtd_read ${2} ${3} | cmp -s ${TMP_IMAGE}
	if [ $? -ne 0 ]; then
		panic "Verifying ${1} image failed!"
	fi

	rm -f ${TMP_IMAGE}
}

# update_volume(file,volname,[len])
update_volume()
{
	local TMP_IMAGE=/tmp/image

	if [ -z "${3}" ]; then
		return
	fi

	echo "Updating ${1} image..."
	zip_extract ${TMP_IMAGE} ${1} ${3}

	uenvfs update ${2} ${TMP_IMAGE}
	if [ $? -ne 0 ]; then
		panic "Updating ${1} image failed!"
	fi

	rm -f ${TMP_IMAGE}
}

# check_uboot(file,[len])
check_uboot()
{
	if [ -z "${2}" ]; then
		return 1
	fi
	if [ ${2} -gt 262144 ]; then
		panic "Update file defective (${1} size)."
	fi
	check_image ${1} 0 ${2}
	return $?
}

# update_uboot(file,[len])
update_uboot()
{
	local TMP_IMAGE=/tmp/image
	local MTD_ERASE=$(get_mtd_erase)
	local SIZE=262144
	local BLOCKS=$(( ( ${SIZE} + ${MTD_ERASE} - 1 ) / ${MTD_ERASE} ))
	local CHKSUM=1

	if [ -z "${2}" ]; then
		return
	fi

	echo "Updating ${1} image..."
	zip_extract ${TMP_IMAGE} ${1} ${2}

	# Check Boot Header of old u-boot.
	# If valid, copy old u-boot to backup.

	for x in $(dd if=/dev/mtd0 bs=4 skip=8 count=11 2>/dev/null | \
	od -An -tx4 -v); do
		CHKSUM=$(( ${CHKSUM} + 0x${x} ))
	done
	CHKSUM=$(( ${CHKSUM} % 0x100000000 ))
	if [ ${CHKSUM} -eq 0 ]; then

		flash_erase -q /dev/mtd0 ${SIZE} ${BLOCKS}
		if [ $? -ne 0 ]; then
			panic "Erasing /dev/mtd0 failed!"
		fi

		dd if=/dev/mtd0 of=/dev/mtd0 bs=${SIZE} count=1 seek=1
		if [ $? -ne 0 ]; then
			panic "Backing up ${1} image failed!"
		fi
	fi

	# Write new u-boot except for Boot Header.
	# Write new Boot Header.

	flash_erase -q /dev/mtd0 0 ${BLOCKS}
	if [ $? -ne 0 ]; then
		panic "Erasing /dev/mtd0 failed!"
	fi

	dd if=${TMP_IMAGE} of=/dev/mtd0 bs=80 skip=1 seek=1
	if [ $? -ne 0 ]; then
		panic "Writing ${1} image failed!"
	fi

	dd if=${TMP_IMAGE} of=/dev/mtd0 bs=80 count=1
	if [ $? -ne 0 ]; then
		panic "Writing ${1} image failed!"
	fi

	rm -f ${TMP_IMAGE}
}

# check_fpgadata(file,[len])
check_fpgadata()
{
	check_md5sum ${1} ${2}
	return 1
}

# update_fpgadata(file,[len])
update_fpgadata()
{
	update_volume ${1} fpgadata ${2}
}

# check_kernel(file,[len])
check_kernel()
{
	check_md5sum ${1} ${2}
	return 1
}

# update_kernel(file,[len])
update_kernel()
{
	update_volume ${1} kernel ${2}
}

# update_sam3u(file,[len])
update_sam3u()
{
	local TMP_IMAGE=/tmp/image

	if [ -z "${2}" ]; then
		return
	fi

	echo "Updating ${1} image..."
	zip_extract ${TMP_IMAGE} ${1} ${2}

	sam3ufw -w ${TMP_IMAGE}
	if [ $? -ne 0 ]; then
		panic "Updating ${1} image failed!"
	fi

	rm -f ${TMP_IMAGE}
}

# update_machxo(file,[len])
update_machxo()
{
	local TMP_IMAGE=/tmp/image

	if [ -z "${2}" ]; then
		return
	fi

	echo "Updating ${1} image..."
	zip_extract ${TMP_IMAGE} ${1} ${2}

	machxofw -w -j ${TMP_IMAGE}
	if [ $? -ne 0 ]; then
		# don't panic ...
		echo "Updating ${1} image failed!" >&2
		echo "Updating ${1} image failed!" >"${STATUSERR}"
	fi

	rm -f ${TMP_IMAGE}
}

export HOME=/tmp
cd /tmp

start_led

if ! mount -o rw -t jffs2 /dev/rw /rw ; then
	panic "Cannot access rw partition!"
fi

# fix old failure messages
rm -rf "${STATUSOLD}.new"
if grep -q 'storage space!\|for partition!' "${STATUSOLD}" 2>/dev/null; then
	# "Update requires at least 10 MiB of free storage space!"
	# "Image ${1} too big for partition!"
	sed -e 's/storage space!/storage space./g' "${STATUSOLD}" \
	    -e 's/for partition!/for partition./g' >"${STATUSOLD}.new"
	mv "${STATUSOLD}.new" "${STATUSOLD}"
	grep -q '!' "${STATUSOLD}" || rm -f "${STATUSOLD}"
fi

# old flash layout: mtd0 256 KiB, mtd1 128 KiB
# new flash layout: mtd0 512 KiB, mtd1 15.5 MiB
if [ $(get_mtd_size 0) -lt 524288 -o $(get_mtd_size 1) -lt 262144 ]; then
	panic "Cannot update old flash layout."
fi

# NRQx old flash: mtd2 32 MiB
# NRQx new flash: mtd2 64 MiB
sensor=$(zip_read ${DATAFS} | tar -xzf - -O .${RSUDESC} | grep SENSOR-TYPE | cut -f2 -d=)
if [ "${sensor}" = "NRQx" -a $(get_mtd_size 2) -lt 67108864 ]; then
	panic "Cannot update ${sensor} with small flash."
fi

# free some storage space ...
rm -rf /rw/${ROOTFS}.new /rw/${OPTFS}.new
rm -rf /rw/etc.new ${SAVE} ${SAVE}.new
if [ "$(df /rw | tail -n 1 | tr -s ' ' | cut -f4 -d' ')" -lt 10240 ] ; then
	panic "Update requires at least 10 MiB of free storage space."
fi

echo "Checking update file..."

SAM3U_LEN=$(zip_length ${SAM3U_IMG})
UBOOT_LEN=$(zip_length ${UBOOT_IMG})
FPGADATA_LEN=$(zip_length ${FPGADATA_IMG})
KERNEL_LEN=$(zip_length ${KERNEL_IMG})
MACHXO_LEN=$(zip_length ${MACHXO_IMG})
ROOTFS_LEN=$(zip_length ${ROOTFS_TARGET})
OPTFS_LEN=$(zip_length ${OPTFS})
DATAFS_LEN=$(zip_length ${DATAFS})
REMOVELIST_LEN=$(zip_length ${REMOVELIST})

if [ "${ROOTFS_LEN}" = "" ] ; then
	panic "Update file defective (${ROOTFS_TARGET} lookup)."
fi

check_md5sum ${SAM3U_IMG} ${SAM3U_LEN}
check_uboot ${UBOOT_IMG} ${UBOOT_LEN} && UBOOT_LEN=
check_fpgadata ${FPGADATA_IMG} ${FPGADATA_LEN} && FPGADATA_LEN=
check_kernel ${KERNEL_IMG} ${KERNEL_LEN} && KERNEL_LEN=
check_md5sum ${MACHXO_IMG} ${MACHXO_LEN}

check_md5sum ${ROOTFS_TARGET} ${ROOTFS_LEN}
check_md5sum ${OPTFS} ${OPTFS_LEN}
check_md5sum ${DATAFS} ${DATAFS_LEN}
check_md5sum ${REMOVELIST} ${REMOVELIST_LEN}

if md5sum_cmp /rw/${ROOTFS} ${ROOTFS_TARGET} ${ROOTFS_LEN} ; then
	ROOTFS_LEN=""
fi

echo "Performing update..."

blink_led

grep -e "^${share_user}:" /rw/etc/passwd | head -n 1 >/rw/etc/passwd.user

tar -cf ${SAVE}.new -C /rw/etc \
 security \
 sysctl.conf \
 network/interfaces \
 passwd.user
mv ${SAVE}.new ${SAVE}
sync

{
grep -e "^${share_user}:" -v /etc/passwd
{
grep -e "^${share_user}:" /rw/etc/passwd
grep -e "^${share_user}:" /etc/passwd
} | head -n 1
} >/tmp/passwd

mv /rw/etc /rw/etc.new
mkdir /tmp/zcip
mv /rw/var/cache/zcip.* /tmp/zcip 2>/dev/null
rm -rf /rw/etc.new /rw/${SUBFS} /rw/devel /rw/testing /rw/opt /rw/startup /rw/var/cache/* 2>/dev/null
mv /tmp/zcip/zcip.* /rw/var/cache/ 2>/dev/null

update_sam3u ${SAM3U_IMG} ${SAM3U_LEN}
update_uboot ${UBOOT_IMG} ${UBOOT_LEN}
update_fpgadata ${FPGADATA_IMG} ${FPGADATA_LEN}
update_kernel ${KERNEL_IMG} ${KERNEL_LEN}
update_machxo ${MACHXO_IMG} ${MACHXO_LEN}

if [ "${REMOVELIST_LEN}" != "" ] ; then
	echo "Removing obsolete files..."
	zip_read ${REMOVELIST} |
	while read l
	do
		eval rm -rf ${l}
	done
fi

if [ "${ROOTFS_LEN}" != "" ] ; then
	echo "Updating root filesystem..."
	zip_extract /rw/${ROOTFS}.new ${ROOTFS_TARGET} ${ROOTFS_LEN}
	sync
	mv -f /rw/${ROOTFS}.new /rw/${ROOTFS}
	sync
fi

(
cd /rw
mount -o loop,ro -t squashfs /rw/${ROOTFS} /ro
cp -a /ro/etc etc.new || panic_space /rw/etc.new
umount /ro
losetup -d /dev/loop0 2>/dev/null
tar -xf ${SAVE} -C etc.new || panic_space /rw/etc.new
cp /tmp/passwd etc.new/passwd || panic_space /rw/etc.new
mv etc.new etc
) >&2
sync
rm -f ${SAVE} /rw/etc/passwd.user

if [ "${OPTFS_LEN}" != "" ] ; then
	echo "Updating application..."
	zip_extract /rw/${OPTFS}.new ${OPTFS} ${OPTFS_LEN}
	sync
	mv -f /rw/${OPTFS}.new /rw/${OPTFS}
	sync
fi

if [ "${DATAFS_LEN}" != "" ] ; then
	echo "Updating application data..."
	zip_read ${DATAFS} | tar -xz -C /rw
fi

# NRPM3: update antennas on reboot
mkdir -p $(dirname ${NRPMALIST})
rm -rf ${NRPMALIST} ${NRPMALIST}.new
printf '%s\n%s\n%s\n' 0 1 2 >${NRPMALIST}.new
sync
mv ${NRPMALIST}.new ${NRPMALIST}

echo "Sync filesystems..."
rm -f "${STATUSOLD}"
rm -f "${STATUSPOW}"
sync
umount /rw

echo "Update done, rebooting system..."
reboot -f

